package gs.log.v0.engine

import gs.log.v0.api.Logger
import cats.effect.Async
import gs.log.v0.data.Log
import gs.log.v0.data.LogLevel
import gs.log.v0.api.Logger.Name
import cats.effect.Ref

import cats.syntax.all.*
import cats.effect.std.Queue
import java.time.Clock
import java.time.Instant

final class QueueLogger[F[_]: Async](
  val name: Logger.Name,
  private val clock: Clock,
  private val maxLevel: Ref[F, LogLevel],
  private val queue: Queue[F, (Log, Log.Metadata)]
) extends Logger[F] {

  override def isLevelEnabled(level: LogLevel): F[Boolean] = 
    maxLevel.get.map(max => level <= max)

  override def warn(log: => Log): F[Unit] = ???

  override def debug(log: => Log): F[Unit] = ???

  override def info(log: => Log): F[Unit] = ???

  override def error(log: => Log): F[Unit] = ???

  override def fatal(log: => Log): F[Unit] = ???

  override def trace(log: => Log): F[Unit] = 
    meta(LogLevel.Trace).flatMap(m => queue.tryOffer((log, m))).flatMap {
      case true => Async[F].unit
      case false => Async[F].unit
    }

  private def now(): F[Instant] =
    Async[F].delay(Instant.now(clock))

  private def meta(level: LogLevel): F[Log.Metadata] =
    now().map(timestamp =>
      Log.Metadata(
        level = level,
        timestamp = timestamp,
        owner = name.show
      )
  )

}
